from __future__ import absolute_import

__description__ = "a module that houses object-based wrappers of keras based classifiers"
__author__ = "Patrick Godwin (patrick.godwin@ligo.org), Kyle Rose (rosek@kenyon.edu)"
__doc__ = "\n\n".join([__description__, __author__])

#-------------------------------------------------

import numpy as np

from idq import configparser
from idq import io
from idq import features
from idq import utils

from . import base as base_classifiers

### required for keras functionality
import tensorflow

from keras.models import Model, Sequential
from keras.layers import Activation, Dense, Dropout, Flatten, Reshape
from keras.layers import LocallyConnected1D as Local1D
from keras.regularizers import l1, l2, l1_l2
from keras_pickle_wrapper import KerasPickleWrapper

### used for class balancing, cross-val
from sklearn.utils import class_weight
from sklearn.model_selection import train_test_split

#-------------------------------------------------

DEFAULT_COLUMN_VALUE = 0

#-------------------------------------------------
# base classes for keras classifiers
#-------------------------------------------------

class KerasModel(base_classifiers.ClassifierModel):
    """the model object for SupervisedKerasClassifier
    """

    def __init__(self, start, end, keras, whitener, channels, select, time=features.DEFAULT_TIME_NAME, segs=None):
        base_classifiers.ClassifierModel.__init__(self, start, end, segs=segs)

        ### define stuff particular to Keras
        self._keras = keras
        self._whitener = whitener
        self._channels = channels
        self._select = select
        self._time = time

    @property
    def keras(self):
        return self._keras()

    @property
    def whitener(self):
        return self._whitener

    @property
    def channels(self):
        return self._channels

    @property
    def select(self):
        return self._select

    @property
    def time(self):
        return self._time

#-------------------------------------------------

class SupervisedKerasClassifier(base_classifiers.SupervisedClassifier):
    """
    A base class for supervised keras classifiers
    Note: not a standalone classifier
    """
    _flavor = "keras_supervised_classifier"
    _required_kwargs = [
        'safe_channels_path',
        'window',
        'time',
        'significance',
        'layers',
        'loss',
        'optimizer',
        'metrics',
        'epochs',
        'batch_size',
    ]

    def __init__(self, *args, **kwargs):
        base_classifiers.SupervisedClassifier.__init__(self, *args, **kwargs)

    def train(self, quiver):
        """
        Trains a model using feature data, makes predictions using the training set's class distribution
        """
        verbose = self.kwargs.get('verbose', False)

        ### set random seed if specified
        random_seed = self.kwargs.get('random_state', None)
        if random_seed is not None:
            if verbose:
                print('setting random seed: %d'%random_seed)
            tensorflow.set_random_seed(random_seed)

        ### create the model, passing in quiver to determine input dimensions
        self.model, data, names = self._define_model(quiver)

        ### balance out class weights if specified
        balanced = self.kwargs.get('balanced', False)
        if balanced:
            if verbose:
                print('balancing class weights for model')
            class_weights = class_weight.compute_class_weight('balanced', np.array([0, 1]), quiver.labels)
        else:
            class_weights = None

        ### train whitener + whiten
        self.model.whitener.train(data, names)
        data = self.model.whitener.whiten(data, names)

        ### if using validation data, shuffle the training set
        if 'validation_split' in self.kwargs:
            train_data, test_data, train_labels, test_labels = train_test_split(
                data,
                quiver.labels,
                test_size=self.kwargs['validation_split'],
                shuffle=True
            )
            validation_data = (test_data, test_labels)
        else:
            train_data, train_labels = data, quiver.labels
            validation_data = None

        ### train model
        self.model.keras.fit(
            train_data,
            train_labels,
            epochs=self.kwargs['epochs'],
            batch_size=self.kwargs['batch_size'],
            class_weight=class_weights,
            validation_data=validation_data,
            verbose=1,
        )

        return self.model

    def evaluate(self, quiver):
        """
        Applies a supervised keras model to feature data.
        """
        if not self.is_trained:
            raise base_classifiers.UntrainedError('%s does not have an internal model'%self.flavor)

        data, names = quiver.vectorize(
            self.model.channels,
            self.model.select,
            time=self.model.time,
            num_proc=self.kwargs.get('num_proc', features.DEFAULT_NUM_PROC),
        )

        ranks=self.model.keras.predict(self.model.whitener.whiten(data, names), verbose=1)
        quiver.apply_ranks(ranks, hashes=self.model.hash)
        return quiver

    def timeseries(self, quiver_factory, dt=base_classifiers.DEFAULT_DT, segs=None):
        """
        Generate a time series of predictions based on predicted model probabilities.
        """
        ### check if model has been trained
        if not self.is_trained:
            raise base_classifiers.UntrainedError('%s does not have an internal model'%self.flavor)

        if segs is None:
            segs = quiver_factory.classifier_data.segs

        ranks = []
        for times in utils.segs2times(segs, dt):
            quiver = quiver_factory(times)
            if len(quiver): ### append ranks if quiver isn't empty
                ranks.append( (self.evaluate(quiver).ranks, times[0], dt) )
        return ranks

    def feature_importance(self):
        """
        return a ranked list of important features within the trained model

        WARNING: NOT IMPLEMENTED YET!
        """
        raise NotImplementedError

    def _define_model(self, quiver):
        """
        Defines and returns a keras classifier model, using keyword arguments as necessary.
        Used internally
        """
        ### set defaults for missing values if not set
        self.kwargs['default'] = self.kwargs.get('default', DEFAULT_COLUMN_VALUE)
        self.kwargs['default_delta_time'] = self.kwargs.get('default_delta_time', -self.kwargs['window'])

        ### vectorize quiver
        channels = io.path2channels(self.kwargs['safe_channels_path'])
        select = features.SelectLoudest(**self.kwargs)
        data, names = quiver.vectorize(
            channels,
            select,
            time=self.kwargs['time'],
            num_proc=self.kwargs.get('num_proc', features.DEFAULT_NUM_PROC),
        )

        ### set quantities used for calculating layer sizes
        self._num_features = len(names)
        self._num_channels = len(channels)
        self._num_columns = self._num_features / self._num_channels

        model = KerasModel(
            quiver.start,
            quiver.end,
            self._define_keras(),
            features.Whitener(),
            channels,
            select,
            time=self.kwargs['time'],
            segs=quiver.segs,
        )

        return model, data, names

    def _define_keras(self, *args, **kwargs):
        return KerasPickleWrapper(Sequential())

    def _add_layer(self, model, layer, input_dim=None, **kwargs):
        """
        given a config-formatted set of layer inputs, generate a Keras layer
        """
        if layer[0] == 'Dense':
            layer_kwargs = {'input_dim': input_dim} if input_dim else {}
            if len(layer) == 5: ### extract regularization params
                _, _, _, reg_type, reg_value = layer
                regularizer = eval(reg_type)(float(reg_value))
                layer_kwargs.update({'kernel_regularizer': regularizer, 'bias_regularizer': regularizer})
            if len(layer) >= 3: ### extract # nodes, activation function
                _, units, activation = layer[:3]
                units = self._process_size(units)
                layer_kwargs.update({'activation': activation})
                model.add(Dense(units, **layer_kwargs))
            else:
                raise ValueError('Dense layer kwargs not formatted correctly: %s'.format(' '.join(layer)))

        elif layer[0] == 'Dropout':
            layer_kwargs = {'input_dim': input_dim} if input_dim else {}
            if len(layer) == 2:
                model.add(Dropout(float(layer[1]), **layer_kwargs))
            else:
                raise ValueError('Dropout layer kwargs not formatted correctly: %s'.format(' '.join(layer)))

        elif layer[0] == 'Local1D':
            layer_kwargs = {}
            ### reshape input layer since Local1D layer expects 3-dim data
            if input_dim:
                model.add(Reshape(input_shape=(input_dim,), target_shape=(input_dim,1)))
            else:
                input_shape = model.layers[-1].output_shape
                if not len(input_shape) == 1:
                    raise ValueError('only 1D input shape allowed for Local1D')
                model.add(Reshape(input_shape=(input_shape[0],), target_shape=(input_shape[0], 1)))

            ### extract layer config
            if len(layer) == 7: ### extract regularization params
                _, _, _, _, _, reg_type, reg_value = layer
                regularizer = eval(reg_type)(float(reg_value))
                layer_kwargs.update({'kernel_regularizer': regularizer, 'bias_regularizer': regularizer})
            if len(layer) >= 5: ### extract locally connected specific params
                _, filters, kernel_size, strides, activation = layer[:5]
                filters, kernel_size, strides = self._process_sizes(filters, kernel_size, strides)
                layer_kwargs.update({'strides': strides, 'activation': activation})
                model.add(Local1D(filters, kernel_size, **layer_kwargs))
                model.add(Flatten())
            else:
                raise ValueError('Local1D layer kwargs not formatted correctly: %s'.format(' '.join(layer)))

        else:
            raise ValueError('layer=%s not understood'%layer[0])

    def _process_sizes(self, *sizes):
        return tuple([self._process_size(size) for size in sizes])

    def _process_size(self, size):
        """
        parse layer sizes from config with special handling for Nchan, Ncol, Ntotal.
        """
        if 'N' in size:
            if 'Ncol' in size:
               size = size.replace('Ncol', '')
               if size:
                   return int(float(size) * self._num_columns)
               else:
                   return self._num_columns
            elif 'Nchan' in size:
               size = size.replace('Nchan', '')
               if size:
                   return int(float(size) * self._num_channels)
               else:
                   return self._num_columns
            elif 'Ntotal' in size:
               size = size.replace('Ntotal', '')
               if size:
                   return int(float(size) * self._num_features)
               else:
                   return self._num_columns
        else:
            return int(size)

#------------------------
#   Neural Networks
#------------------------

class NeuralNetwork(SupervisedKerasClassifier):
    """
    A neural network based on Keras
    """
    _flavor = "neural_network"

    def _define_keras(self, **kwargs):
        model = Sequential()

        ### generate layers sequentially
        layers = configparser.config2layers(self.kwargs['layers'])
        self._add_layer(model, layers[0], input_dim=self._num_features)
        for layer in layers[1:]:
            self._add_layer(model, layer)

        ### compile the model
        metrics = self.kwargs['metrics'].strip().split()
        model.compile(loss=self.kwargs['loss'], optimizer=self.kwargs['optimizer'], metrics=metrics)

        return KerasPickleWrapper(model)
